; II užduotis
; Mini disasembleris
; Parašyti programą, kuri nuskaito iš pateikto .com failo komandas ir jas 
; disasembliuoja. Failo pavadinimas gali būti fiksuotas. Com faile bus ne 
; daugiau dešimties komandų. Komandos bus tik MOV, JNC, DIV


;******************************************************************************
; MOV – Move Data
; register1 to register2		1000 100w : 11 reg1 reg2
; register2 to register1		1000 101w : 11 reg1 reg2
; memory to reg					1000 101w : mod reg r/m
; reg to memory					1000 100w : mod reg r/m
; immediate to register			1100 011w : 11 000 reg : immediate data
; immediate to register 
;	(alternate encoding) 		1011 w reg : immediate data
; immediate to memory			1100 011w : mod 000 r/m : immediate data
; memory to AL, AX, or EAX 		1010 000w : full displacement
; AL, AX, or EAX to memory 		1010 001w : full displacement

; MOV – Move to/from Control Registers
; CR0 from register 			0000 1111 : 0010 0010 : 11 000 reg
; CR2 from register 			0000 1111 : 0010 0010 : 11 010 reg
; CR3 from register 			0000 1111 : 0010 0010 : 11 011 reg
; CR4 from register 			0000 1111 : 0010 0010 : 11 100 reg
; register from CR0-CR4 		0000 1111 : 0010 0000 : 11 eee reg

; MOV – Move to/from Debug Registers
; DR0-DR3 from register 		0000 1111 : 0010 0011 : 11 eee reg
; DR4-DR5 from register 		0000 1111 : 0010 0011 : 11 eee reg
; DR6-DR7 from register 		0000 1111 : 0010 0011 : 11 eee reg
; register from DR6-DR7 		0000 1111 : 0010 0001 : 11 eee reg
; register from DR4-DR5 		0000 1111 : 0010 0001 : 11 eee reg
; register from DR0-DR3 		0000 1111 : 0010 0001 : 11 eee reg

; MOV – Move to/from Segment Registers
; register to segment register 	1000 1110 : 11 sreg3 reg
; register to SS 				1000 1110 : 11 sreg3 reg
; memory to segment reg 		1000 1110 : mod sreg3 r/m
; memory to SS 					1000 1110 : mod sreg3 r/m
; segment register to register 	1000 1100 : 11 sreg3 reg
; segment register to memory 	1000 1100 : mod sreg3 r/m

; Jcc – Jump if Condition is Met
; 8-bit displacement 			0111 tttn : 8-bit displacement
; full displacement 			0000 1111 : 1000 tttn : full displacement

; DIV – Unsigned Divide
; AL, AX, or EAX by register 	1111 011w : 11 110 reg
; AL, AX, or EAX by memory 		1111 011w : mod 110 r/m

; w - Specifies if data is byte or full-sized, where full-sized is either 
;	16 or 32 bits (refer to Table B-4)
; tttn - For conditional instructions, specifies a condition asserted or a 
;	condition negated (refer to Table B-8)
;******************************************************************************

;******************************************************************************
; MOV – MOVe – persiuntimas
;
; 1. registras <=> registras / atmintis;
; 1000 10dw mod reg r/m [poslinkis];
; d = 0 : reg => r / m
; d = 1 : r / m <= reg
;
; 2. betarpiškas operandas => registras / atmintis;
; 1100 011w mod 000 r/m [poslinkis] bet.op1 [bet.op.2, jei w = 1]
; Krypties bito nėra, nes kryptis jau yra apibrėžta.
; Laukas reg nereikalingas, jis yra naudojamas kaip operacijos kodo išplėtimas.
; Poslinkis suformuojamas priklausomai nuo lauko mod reikšmės.
;
; 3. betarpiškas operandas => registras
; 1011 w reg bet.op.1 [bet.op.2, jei w = 1]
;
; 4. atmintis => akumuliatorius
; 1010 000w adr.j.b adr.j.b adr.v.b.
; Komandos formatas yra toks, kad registras ir operandas betarpiškai nenurodyti.
; Iš operacijos kodo yra žinoma, kad registras yra akumuliatorius.
;
; 5. akumuliatorius => atmintis
; 1010 001w adr.j.b adr.j.b adr.v.b.
; Pastaba. Atvejų 4. ir 5. negalima sujungti į vieną 101000 d w adr.j.b adr.j.b 
; adr.v.b., nes bitas d apibėžia priešingos krypties perdavimą, negu yra 
; nurodyta 4. ir 5. atvejais.
;
; 6. registras / atmintis <=> segmento registras
; 1000 11d0 mod 0 sreg r/m [poslinkis]
; sreg negali būti lygus 01, nes negalima nusiųsti reikšmės į registrą CS, nes 
; priešingu atveju priskyrimo veiksmu pakeistume CS registro reikšmę, o ji gali 
; būti pakeista tik valdymo perdavimo komandomis.
; Nėra w bito, nes operandas visada 2 baitų.
;
;
; DIV – DIVide – skaičių be ženklo dalyba.
;
; 1111 011w mod 110 r/m [poslinkis]
; Jeigu w = 0, tada daliklis yra baitas, ir:
; registre AX esanti reikšmė : operandas 2 	=> dalmuo registre AL
; 											=> liekana registre AH
; Jeigu w = 1, tada daliklis yra žodis, ir:
; registruose[DX, AX] esanti reikšmė : operandas 2 	=> dalmuo registre AX
; 													=> liekana registre DX
; Jeigu dalmuo netelpa registre AL (kai w = 0) arba AX (kai w = 1), tada
; generuojamas pertraukimas – dalyba iš nulio.
;
;
; JNC - Jump if Not Carry
;
; formatas: op.kodas + poslinkis
; op.kodas = 73
;******************************************************************************


; pertraukimų parametrai
OPEN equ 3dh			; failo atidarymas
CREATE equ 3ch			; failo sukūrimas
SEEK equ 42h			; failo rodyklės perkėlimas
READ equ 3fh			; failo skaitymas
WRITE equ 40h			; failo rašymas
CLOSE equ 3eh			; failo uždarymas

; kaukės
BIT0 equ 00000001B
BIT1 equ 00000010B
BIT2 equ 00000100B
BIT3 equ 00001000B
BIT10 equ 00000011B
; ModR/M baito kaukės
MOD_BIT equ 11000000B
REG_BIT equ 00111000B
RM_BIT equ 00000111B
; mod bitų reikšmės
OP_M0 equ 00000000B		; atmintis, be postūmio
OP_M1 equ 01000000B		; atmintis, postūmis 1
OP_M2 equ 10000000B		; atmintis, postūmis 2
OP_REG equ 11000000B		; registras

;******************************************************************************
; STEKO SEGMENTAS
;******************************************************************************
SSeg segment para stack 'stack'
	db 100h dup (?)
SSeg ends
;******************************************************************************

;******************************************************************************
; DUOMENŲ SEGMENTAS
;******************************************************************************
DSeg segment
	; failų duomenys
	com_failo_vardas db "asm2com.com", 0
	rez_failo_vardas db "asm2.rez", 0
	com_failas dw ?
	rez_failas dw ?
	
	; buferių duomenys
	buf_dyd dw ?		; kiek baitų laikoma failo buferyje
	buf db 60 dup (?)	; failo buferis <=10 komandų po <=6 baitus
	rbuf db 255 dup (?)	; rezultato buferis
	
	daliklis dw 0ah
	
	komandu_kiekis dw 10	
	
	;**************************************************************************
	; komandų kaukės:
	; xor kaukės naudojamos atpažinti komandai
	; and kaukės naudojamos nunulint nereikalingus bitus
	;**************************************************************************
CMD_KAUKES:
	div_cmd_and db 11111110B
	div_cmd_xor db 11110110B
	
	mov_cmd1_and db 11111100B
	mov_cmd1_xor db 10001000B
	
	mov_cmd2_and db 11111110B
	mov_cmd2_xor db 11000110B
	
	mov_cmd3_and db 11110000B
	mov_cmd3_xor db 10110000B
	
	mov_cmd4_and db 11111111B
	mov_cmd4_xor db 00001111B
	
	mov_cmd5_and db 11111100B
	mov_cmd5_xor db 10100000B
	
	mov_cmd6_and db 11111100B
	mov_cmd6_xor db 10001100B
	
	jnc_cmd1_and db 11111111B
	jnc_cmd1_xor db 01110011B
	
	jnc_cmd2_and db 11111111B
	jnc_cmd2_xor db 00001111B
	
	nop_cmd_and db 11111111B
	nop_cmd_xor db 10010000B
	
	;**************************************************************************
	
	; komandų prefiksai
	prefiksu_kiekis db 6
PREFIKSAI:
	cs_override_pref db 2eh
	ss_override_pref db 36h
	ds_override_pref db 3eh
	es_override_pref db 26h
	fs_override_pref db 64h
	gs_override_pref db 65h
	
	prefiksu_lent dw cs_eil, ss_eil, ds_eil, es_eil, fs_eil, gs_eil
	
	; komandų nagrinėjimo procedūrų adresai
CMD_PROC:
	div_proc dw NAGR_DIV
	mov_proc1 dw NAGR_MOV1
	mov_proc2 dw NAGR_MOV2
	mov_proc3 dw NAGR_MOV3
	mov_proc4 dw NAGR_MOV4
	mov_proc5 dw NAGR_MOV5
	mov_proc6 dw NAGR_MOV6
	jnc_proc1 dw NAGR_JNC1
	jnc_proc2 dw NAGR_JNC2
	nop_proc dw NAGR_NOP
	
	; registrų lentelės (pavadinimų adresai)
	duom_reg8 dw al_eil, cl_eil, dl_eil, bl_eil, ah_eil, ch_eil, dh_eil, bh_eil
	duom_reg16 dw ax_eil, cx_eil, dx_eil, bx_eil, sp_eil, bp_eil, si_eil, di_eil
	creg dw cr0_eil, 0, cr2_eil, cr3_eil, cr4_eil
	dreg dw dr0_eil, dr1_eil, dr2_eil, dr3_eil, 0, 0, dr6_eil, dr7_eil
	sreg dw es_eil, cs_eil, ss_eil, ds_eil
	
	; adresacijos lentelė
	adr_lent dw mem_000, mem_001, mem_010, mem_011, mem_100, mem_101, mem_110, mem_111
	
	;**************************************************************************
	; eilutės (pirmas baitas nurodo eilutės ilgį)
	;**************************************************************************
	
	; komandų vardai
	div_eil db 4, "div "
	jnc_eil db 4, "jnc "
	mov_eil db 4, "mov "
	nop_eil db 4, "nop", 13
	
	; registrų vardai
	al_eil db 2, "al"
	ah_eil db 2, "ah"
	bl_eil db 2, "bl"
	bh_eil db 2, "bh"
	cl_eil db 2, "cl"
	ch_eil db 2, "ch"
	dl_eil db 2, "dl"
	dh_eil db 2, "dh"
	ax_eil db 2, "ax"
	bx_eil db 2, "bx"
	cx_eil db 2, "cx"
	dx_eil db 2, "dx"
	sp_eil db 2, "sp"
	bp_eil db 2, "bp"
	si_eil db 2, "si"
	di_eil db 2, "di"
	
	cr0_eil db 3, "cr0"
	cr2_eil db 3, "cr2"
	cr3_eil db 3, "cr3"
	cr4_eil db 3, "cr4"
	
	dr0_eil db 3, "dr0"
	dr1_eil db 3, "dr1"
	dr2_eil db 3, "dr2"
	dr3_eil db 3, "dr3"
	dr6_eil db 3, "dr6"
	dr7_eil db 3, "dr7"
	
	cs_eil db 2, "cs"
	ss_eil db 2, "ss"
	ds_eil db 2, "ds"
	es_eil db 2, "es"
	fs_eil db 2, "fs"
	gs_eil db 2, "gs"
	
	byteptr_eil db 9, "byte ptr "

	; adresacijos eilutės
	mem_000 db 5, "bx+si"
	mem_001 db 5, "bx+di"
	mem_010 db 5, "bp+si"
	mem_011 db 5, "bp+di"
	mem_100 db 2, "si"
	mem_101 db 2, "di"
	mem_110 db 2, "bp"
	mem_111 db 2, "bx"
	
	;**************************************************************************
	
DSeg ends
;******************************************************************************

;******************************************************************************
; KODO SEGMENTAS
;******************************************************************************
CSeg segment
	assume cs:CSeg, ds:DSeg, ss:SSeg
	
	;**************************************************************************
	; pagrindinė procedūra MAIN
	;**************************************************************************
	MAIN proc far
		; Inicializuojam programos įėjimo adresą
		push	ds
		sub		ax, ax
		push	ax
		mov		ax, DSeg
		mov		ds, ax
		mov		es, ax
		;push	bx

		; Programos kodas
		
		; atidarome com failą
		mov		ah, OPEN
		mov		al, 0				; read only
		mov		dx, offset com_failo_vardas
		int		21h
		jnc		TESTI_MAIN
		ret							; jei klaida - išeiti
	TESTI_MAIN:
		mov		com_failas, ax		; failo handle patalpinama į com_failas		
		
		; skaitome i buferį
		mov		ah, READ
		mov		bx, com_failas
		mov		cx, 60				; bandom skaityti 60 baitų
		mov		dx, offset buf
		int		21h
		mov		buf_dyd, ax			; ax - nuskaitytų baitų skaičius
		
		; uždaromas com failas
		mov		ah, CLOSE
		mov		bx, com_failas
		int		21h
		
		
		; nagrinėjame buferį
		
		mov		si, offset buf
		mov		di, offset rbuf
	MAIN_CIKLAS:
		mov		ax, offset buf
		add		ax, buf_dyd
		sub		ax, si
		jz		IRASYTI				; baigėsi buferis - įrašom viską į failą
		
		xor		ax, ax
		; tikriname, ar ne prefiksas
		call	TIKRINTI_PREF
		
		; tikriname kiekvieną komandą, kol randam tinkančią
		mov		cx, komandu_kiekis
		mov		dx, offset CMD_KAUKES
		mov		bx, offset CMD_PROC
	MAIN_CIKLAS2:
		
		mov		al, byte ptr [si]
		push	si
		mov		si, dx
		and		al, byte ptr [si]
		xor		al, byte ptr [si+1]
		pop		si
		jz		NAGR_KOMANDA
		
	TESTI_MAIN_CIKLAS2:
		dec		cx
		jz		BAIGTI				; mus apgavo - tokios komandos nėra! išeinam
		add		dx, 2
		add		bx, 2
		jmp		MAIN_CIKLAS2
		
	NAGR_KOMANDA:
		call	[bx]
		cmp		ax, 0
		jne		TESTI_MAIN_CIKLAS2	; ne ta komanda. bandom toliau
		jmp		MAIN_CIKLAS
	
		
		; rezultatai įrašomi į failą ir išeinama
	
	IRASYTI:
		; sukuriame rezultatų failą
		mov		ah, CREATE
		mov		cx, 22h
		mov		dx, offset rez_failo_vardas
		int		21h
		jc		BAIGTI				; jei handle nenustatyta - išeiti
		mov		rez_failas, ax		; failo handle patalpinama į rez_failas
		
		; rašome rezultatų buferio turinį
		mov		ah, WRITE
		mov		bx, rez_failas
		mov		cx, di
		sub		cx, offset rbuf
		mov		dx, offset rbuf
		int		21h
		
		; uždaromas rezultatų failas
		mov		ah, CLOSE
		mov		bx, rez_failas
		int		21h

		; Grįžti į DOS
	BAIGTI:
		ret
	MAIN endp
	
	;**************************************************************************
	; procedūra TIKRINTI_PREF - tikrina, ar prieš komandą yra prefiksų
	;	jei yra, jų skaičius įrašomas į ah 2 ir 3 bitus (kol kas gali
	;	būti tik 01)
	;	si rodo į buferį
	;**************************************************************************
	TIKRINTI_PREF proc near
		push	cx
		push	di
		
		; pradinė 2 ir 3 bitų reikšmė
		and		ah, 11110011B
		
	PREF_CIKLAS1:
		mov		cl, prefiksu_kiekis
		mov		di, offset PREFIKSAI
	PREF_CIKLAS2:
		mov		ch, byte ptr[di]
		cmp		ch, byte ptr[si]
		je		NUSTAT_PREFIKSA
		inc		di
		dec		cl
		jz		BAIGTI_TIKRINTI_PREF
		jmp		PREF_CIKLAS2
	NUSTAT_PREFIKSA:
		or		ah, BIT2
		inc		si
		jmp		PREF_CIKLAS1
		
	BAIGTI_TIKRINTI_PREF:
		pop		di
		pop		cx
		ret
	TIKRINTI_PREF endp
	
	;**************************************************************************
	; procedūra NAGR_DIV - nagrinėja div komandą:
	;	ah saugoma papildoma informacija (0 bitas - w, 1 bitas - d,
	;	2 ir 3 bitai - prefiksų kiekis, 4,5 ir 6 - ModR/M baito reg lauko
	;	paskirtis: 000 - komandos išplėtimas, 001 - duomenų registras,
	;	010 - control registras, 011 - debug registras 100 - segmento registras;
	;	kviečiant NAGR_DIV būna užpildyti tik 2 ir 3 bitai)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_DIV proc near
		push	bx
		
		; ar tikrai div?
		mov		al, byte ptr [si+1]
		and		al, REG_BIT
		cmp		al, 00110000B
		je		NAGR_DIV_TOLIAU
		mov		ax, 1
		jmp		BAIGTI_NAGR_DIV
		
	NAGR_DIV_TOLIAU:
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset div_eil
		call	RASYTI_EIL
		pop		si
		
		; nagrinėjame tolesnį komandos baitą - ModR/M
		and		ah, 10001100B
		mov		al, byte ptr[si]
		and		al, BIT10
		or		ah, al
		inc		si
		call	NAGR_MODRM
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
	BAIGTI_NAGR_DIV:
		pop		bx
		ret
	NAGR_DIV endp
	
	;**************************************************************************
	; procedūra NAGR_MOV1 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV1 proc near
		push	bx
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		; nagrinėjame tolesnį komandos baitą - ModR/M
		mov		al, byte ptr[si]
		and		ah, 10001100B
		and		al, BIT10
		or		ah, al
		or		ah, 00010000B
		inc		si
		call	NAGR_MODRM
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
	BAIGTI_NAGR_MOV1:
		pop		bx
		ret
	NAGR_MOV1 endp
	
	;**************************************************************************
	; procedūra NAGR_MOV2 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV2 proc near
		push	bx
		
		; ar tikrai mov?
		mov		al, byte ptr [si+1]
		and		al, 00111000B
		jnz		BAIGTI_NAGR_MOV2
	
	NAGR_MOV2_TOLIAU:
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		; nagrinėjame tolesnį komandos baitą - ModR/M
		and		ah, 10001100B
		mov		al, byte ptr[si]
		and		al, BIT10
		or		ah, al
		inc		si
		call	NAGR_MODRM
		inc		si
		
		mov		byte ptr [di], ','
		inc		di
		
		; rašome antrą parametrą
		and		ah, BIT0
		jz		NAGR_MOV2_8BIT
		
		call	RASYTI_SK_2B
		inc		si
		jmp		NAGR_MOV2_TOLIAU2
		
	NAGR_MOV2_8BIT:
		call	RASYTI_SK_1B
		
	NAGR_MOV2_TOLIAU2:		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
	BAIGTI_NAGR_MOV2:
		pop		bx
		ret
	NAGR_MOV2 endp
	
	;**************************************************************************
	; procedūra NAGR_MOV3 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV3 proc near
		push	bx
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		and		ah, 10001100B
		mov		al, byte ptr [si]
		inc		si
		mov		bl, al
		and		bl, BIT3
		jz		NAGR_MOV3_TESTI
		or		ah, BIT0
		
	NAGR_MOV3_TESTI:
		; nagrinėjame registrą
		mov		bl, al
		and		bl, 00000111B
		mov		bh, 00010000B
		call	NAGR_REG
		
		; dedame kablelį
		mov		byte ptr [di], ','
		inc		di
		
		; nagrinėjame skaičių
		and		ah, BIT0
		jnz		NAGR_MOV3_SK2B
		
		call	RASYTI_SK_1B
		jmp		BAIGTI_NAGR_MOV3
	
	NAGR_MOV3_SK2B:
		call	RASYTI_SK_2B
		inc		si
		
	BAIGTI_NAGR_MOV3:
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
		pop		bx
		ret
	NAGR_MOV3 endp
	
	;**************************************************************************
	; procedūra NAGR_MOV4 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV4 proc near
		push	bx
		
		; ar tikrai mov?
		mov		al, byte ptr [si+1]
		and		al, 11111100B
		xor		al, 00100000B
		jnz		BAIGTI_NAGR_MOV2
		mov		al, byte ptr [si+2]
		and		al, 11000000B
		xor		al, 11000000B
		jnz		BAIGTI_NAGR_MOV2
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		; nagrinėjame tolesnį komandos baitą - ModR/M
		inc		si
		mov		al, byte ptr[si]
		and		ah, 10001100B
		and		al, BIT10
		or		ah, al
		or		al, BIT1
		shl		al, 4
		or		ah, al
		inc		si
		call	NAGR_MODRM
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
	BAIGTI_NAGR_MOV4:
		pop		bx
		ret
	NAGR_MOV4 endp
	
	;**************************************************************************
	; procedūra NAGR_MOV5 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV5 proc near
		push	bx
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		; nustatome kryptį
		mov		al, byte ptr [si]
		and		al, BIT1
		jz		MOV5_IS_ATM
		
		; mov [adr], reg
		inc		si
		call	NAGR_PREF
		mov		byte ptr [di], '['
		inc		di
		call	RASYTI_SK_2B
		mov		byte ptr [di], ']'
		inc		di
		mov		byte ptr [di], ','
		inc		di
		
		; nustatome operandų dydį
		mov		al, byte ptr [si-1]
		and		al, BIT0
		jz		MOV5_IS_REG_1B
		
		push	si
		mov		si, offset ax_eil
		call	RASYTI_EIL
		pop		si
		jmp		MOV5_TESTI
		
	MOV5_IS_REG_1B:
		push	si
		mov		si, offset al_eil
		call	RASYTI_EIL
		pop		si
		jmp		MOV5_TESTI
		
	MOV5_IS_ATM:
		; mov reg, [adr]
		
		; nustatome operandų dydį
		mov		al, byte ptr [si]
		and		al, BIT0
		inc		si
		jz		MOV5_IS_ATM_1B
		
		push	si
		mov		si, offset ax_eil
		call	RASYTI_EIL
		pop		si
		jmp		MOV5_IS_ATM_TESTI
		
	MOV5_IS_ATM_1B:
		push	si
		mov		si, offset al_eil
		call	RASYTI_EIL
		pop		si
		
	MOV5_IS_ATM_TESTI:
		mov		byte ptr [di], ','
		inc		di
		call	NAGR_PREF
		mov		byte ptr [di], '['
		inc		di
		call	RASYTI_SK_2B
		mov		byte ptr [di], ']'
		inc		di
	
	MOV5_TESTI:
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		add		si, 2
		
		pop		bx
		ret
	NAGR_MOV5 endp
	
	;**************************************************************************
	; procedūra NAGR_MOV6 - nagrinėja mov komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_MOV6 proc near
		push	bx
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset mov_eil
		call	RASYTI_EIL
		pop		si
		
		; nagrinėjame tolesnį komandos baitą - ModR/M
		mov		al, byte ptr[si]
		and		ah, 10001100B
		and		al, BIT1
		or		ah, al
		or		ah, 01000001B
		inc		si
		call	NAGR_MODRM
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
	BAIGTI_NAGR_MOV6:
		pop		bx
		ret
	NAGR_MOV6 endp
	
	;**************************************************************************
	; procedūra NAGR_JNC1 - nagrinėja jnc komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_JNC1 proc near
		push	bx
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset jnc_eil
		call	RASYTI_EIL
		pop		si
		
		inc		si
		call	RASYTI_SK_1B_ZENKL
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		inc		si
		
		pop		bx
		ret
	NAGR_JNC1 endp
	
	;**************************************************************************
	; procedūra NAGR_JNC2 - nagrinėja jnc komandą:
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_JNC2 proc near
		push	bx
		
		; ar tikrai jnc?
		inc		si
		mov		al, byte ptr [si]
		xor		al, 10000011B
		jnz		BAIGTI_NAGR_JNC2
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset jnc_eil
		call	RASYTI_EIL
		pop		si
		
		inc		si
		call	RASYTI_SK_2B
		
		; baigiame eilutę buferyje
		mov		byte ptr[di], 13
		inc		di
		
		; viskas gerai
		mov		ax, 0
		add		si, 2
		
	BAIGTI_NAGR_JNC2:
		pop		bx
		ret
	NAGR_JNC2 endp
	
	;**************************************************************************
	; procedūra NAGR_NOP - nagrinėja nop komandą (no operation):
	;	ah saugoma papildoma informacija (žr. NAGR_DIV)
	;	si - pirmo komandos baito adresas, di - rezultatų buferio rodyklė
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;**************************************************************************
	NAGR_NOP proc near
		
		; įrašome į buferį komandos pavadinimą
		push	si
		mov		si, offset nop_eil
		call	RASYTI_EIL
		pop		si
		
		xor		ax, ax
		inc		si
		ret
	NAGR_NOP endp
	
	;**************************************************************************
	; procedūra NAGR_MODRM nagrinėja ModR/M baitą:
	;	ah saugo papildomą informaciją (žr. NAGR_DIV),
	;	si - nagrinėjamo baito adresą,
	;	di - rezultatų buferio rodyklę
	;	išeinant ax = 0, jei viskas gerai, 1 - jei klaida
	;	si perstumiamas už komandos
	;	keičiamas al, bx
	;**************************************************************************
	NAGR_MODRM proc near
		mov		al, ah
		and		al, BIT1
		jz		PIRM_OP_RM		; pirmas operandas r/m
	
		; pirmas operandas - reg
	PIRM_OP_REG:
		mov		al, ah
		and		al, 01110000B
		jz		ANTR_OP_RM
		mov		bl, byte ptr[si]
		and		bl, REG_BIT
		shr		bl, 3
		mov		bh, ah
		call	NAGR_REG
		mov		byte ptr[di], ','
		inc		di
		
	ANTR_OP_RM:
		call	NAGR_RM
		jmp		BAIGTI_NAGR_MODRM
		
		; pirmas operandas - r/m
	PIRM_OP_RM:
		mov		bl, byte ptr[si]
		call	NAGR_RM
		mov		byte ptr[di], ','
		inc		di
	ANTR_OP_REG:
		and		bl, REG_BIT
		shr		bl, 3
		mov		bh, ah
		call	NAGR_REG
		
	BAIGTI_NAGR_MODRM:
		ret
	NAGR_MODRM endp
	
	
	;**************************************************************************
	; procedūra NAGR_REG nagrinėja registro kodą:
	;	ah saugo papildomą informaciją (žr. NAGR_DIV), 
	;	bl - registro kodą, bh 4 ir 5 bitai - registrų rūšį,
	;	di - rezultatų buferio adresą
	;	keičiamas al
	;**************************************************************************
	NAGR_REG proc near
		push	si
		push	bx		
		
		; nustatom su kokiais registrais dirbam
		and		bh, 01110000B
		jz		BAIGTI_NAGR_REG
		cmp		bh, 00100000B
		je		NAGR_CR
		cmp		bh, 00110000B
		je		NAGR_DR
		cmp		bh, 01000000B
		je		NAGR_SREG
		
		; nustatom su kokio dydžio registrais dirbam
		mov		al, ah
		and		al, BIT0
		jz		NAGR_REG8
		mov		si, offset duom_reg16
		jmp		TOLIAU_NAGR_REG
	NAGR_REG8:
		mov		si, offset duom_reg8
		jmp		TOLIAU_NAGR_REG
	
	NAGR_CR:
		mov		si, offset creg
		jmp		TOLIAU_NAGR_REG
	
	NAGR_DR:
		mov		si, offset dreg
		jmp		TOLIAU_NAGR_REG
	
	NAGR_SREG:
		mov		si, offset sreg
		
	TOLIAU_NAGR_REG:
		; nustatome registro vardo eilutės adresą
		xor		bh, bh
		add		si, bx
		add		si, bx
		mov		si, [si]
		
		; rašome registro vardą į buferį
		call	RASYTI_EIL
	
	BAIGTI_NAGR_REG:
		pop		bx
		pop		si
		ret
	NAGR_REG endp
	
	
	;**************************************************************************
	; procedūra NAGR_RM - nagrinėja r/m operandą
	;	ah saugo papildomą informaciją (žr. NAGR_DIV),
	;	si - ModRM baito adresą, di - buferio adresą
	;**************************************************************************
	NAGR_RM proc near
		push	ax
		push	bx
	
		mov		bl, byte ptr [si]
		mov		bh, bl
		
		; išskiriame mod bitus
		and		bl, MOD_BIT
		
		; tikriname, kokia mod reikšmė
		cmp		bl, OP_REG
		jne		NAGR_OP_M
		
		; operandas - registras
		mov		bl, bh
		and		bl, RM_BIT
		mov		bh, 00010000B
		call	NAGR_REG
		jmp		BAIGTI_RM		; išeinam
		
	NAGR_OP_M:
		; jei reikia, pridedam "byte ptr"
		mov		al, ah
		and		al, BIT0
		jnz		NAGR_RM_TOLIAU1
		push	si
		mov		si, offset byteptr_eil
		call	RASYTI_EIL
		pop		si
		
	NAGR_RM_TOLIAU1:		
		; žiūrim ar buvo prefiksas
		call	NAGR_PREF
	
		; nustatome operando adresą atmintyje
		mov		byte ptr [di], "["
		inc		di
		push	si
		push	cx
		xor		ch, ch
		mov		cl, bh
		and		cl, RM_BIT
		mov		si, offset adr_lent
		add		si, cx
		add		si, cx
		mov		si, [si]		; si rodo į reikalingą adresacijos eilutę
		pop		cx
		
		cmp		bl, OP_M0
		je		NAGR_OP_M0		; operandas - atmintis, poslinkio nėra
		cmp		bl, OP_M1
		je		NAGR_OP_M1		; operandas - atmintis, poslinkis - 1b
		
		; operandas - atmintis, poslinkis - 2b
		call	RASYTI_EIL
		pop		si
		mov		byte ptr [di], "+"
		inc		si
		call	RASYTI_SK_2B
		inc		si
		jmp		NAGR_RM_TOLIAU3
	
	NAGR_OP_M0:
		; operandas - atmintis, poslinkio nėra
		and		bh, RM_BIT
		cmp		bh, 110B
		je		NAGR_OP_M0_110	; atskiras atvejis, kai yra tik poslinkis (2b)
		call	RASYTI_EIL
		pop		si
		jmp		NAGR_RM_TOLIAU3
	
	NAGR_OP_M0_110:
		pop		si
		inc		si
		call	RASYTI_SK_2B
		inc		si
		jmp		NAGR_RM_TOLIAU3
		
	NAGR_OP_M1:
		call	RASYTI_EIL
		pop		si
		inc		si
		call	RASYTI_SK_1B_ZENKL
	
	NAGR_RM_TOLIAU3:
		mov		byte ptr [di], "]"
		inc		di
		
	BAIGTI_RM:
		pop		bx
		pop		ax
		ret
	NAGR_RM endp
	
	;**************************************************************************
	; procedūra NAGR_PREF - nagrinėja segment override prefiksus
	;	ah - papildomi duomenys (žr. NAGR_DIV)
	;	si - antras komandos baitas
	;	di - buferio adresas
	;	keičiamas al
	;**************************************************************************
	NAGR_PREF proc near
		push	si
		push	bx
		push	ax
	
		; tikrinam, ar yra prefiksas
		mov		al, ah
		and		al, BIT2
		jz		BAIGTI_NAGR_PREF
		
		; randam prefiksą
		mov		al, byte ptr[si-2]
		mov		bx, offset PREFIKSAI
		mov		si, offset prefiksu_lent
	NAGR_PREF_CIKLAS:
		cmp		byte ptr [bx], al
		je		BAIGTI_NAGR_PREF_CIKLAS
		inc		bx
		add		si, 2
		jmp		NAGR_PREF_CIKLAS
		
	BAIGTI_NAGR_PREF_CIKLAS:
		; rašome prefikso informaciją (segmento vardą)
		mov		si, [si]
		call	RASYTI_EIL
		mov		byte ptr [di], ":"
		inc		di
		
	BAIGTI_NAGR_PREF:
		pop		ax
		pop		bx
		pop		si
		ret
	NAGR_PREF endp
	
	;**************************************************************************
	; procedūra RASYTI_EIL - į rezultatų buferį rašo eilutę
	;	si - eilutės, kurios pirmas baitas nurodo jos ilgį, adresas
	;	di - buferio adresas
	;**************************************************************************
	RASYTI_EIL proc near
		push	cx
		xor		cx, cx
		mov		cl, byte ptr[si]
		inc		si
		rep		movs	byte ptr es:[di], byte ptr ds:[si]
		pop		cx
		ret
	RASYTI_EIL endp
	
	;**************************************************************************
	; procedūra RASYTI_SK_1B_ZENKL - į rezultatų buferį rašo 1b dydžio skaičių
	;	su ženklu
	;	si - skaičiaus adresas, di - buferio adresas
	;**************************************************************************
	RASYTI_SK_1B_ZENKL proc near
		push	ax
		
		mov		al, byte ptr [si]
		cmp		al, 0
		jl		RASYTI_MINUS
		mov		byte ptr [di], "+"
		jmp		RASYTI_SK_1B_TESTI
	RASYTI_MINUS:
		mov		byte ptr [di], "-"
		; padarome skaičių teigiamu
		xor		ah, ah
		sub		ah, al
		mov		al, ah
		
	RASYTI_SK_1B_TESTI:
		inc		di
		xor		ah, ah
		call	RASYTI_SK
		
		pop		ax
		ret
	RASYTI_SK_1B_ZENKL endp
	
	;**************************************************************************
	; procedūra RASYTI_SK_1B - į rezultatų buferį rašo 1b dydžio skaičių
	;	be ženklo
	;	si - skaičiaus adresas, di - buferio adresas
	;**************************************************************************
	RASYTI_SK_1B proc near
		push	ax
		xor		ax, ax
		mov		al, byte ptr [si]
		call	RASYTI_SK
		pop		ax
		ret
	RASYTI_SK_1B endp
	
	;**************************************************************************
	; procedūra RASYTI_SK_2B - į rezultatų buferį rašo 2b dydžio skaičių
	;	be ženklo
	;	si - skaičiaus adresas, di - buferio adresas
	;**************************************************************************
	RASYTI_SK_2B proc near
		push	ax
		mov		ax, [si]
		call	RASYTI_SK
		pop		ax
		ret
	RASYTI_SK_2B endp
	
	;**************************************************************************
	; procedūra RASYTI_SK - į rezultatų buferį rašo skaičių be ženklo
	;	ax - skaičius, di - buferio adresas
	;**************************************************************************
	RASYTI_SK proc near
		push	cx
		push	dx
		
		; jei iš pradžių nulis - rašom '0' ir išeinam
		cmp		ax, 0
		jne		RASYTI_SK_TESTI
		mov		byte ptr [di], "0"
		inc		di
		jmp		BAIGTI_RASYTI_SK
		
	RASYTI_SK_TESTI:
		; skaičiaus skaitmenys sukeliami į steką
		xor		cx, cx
	RASYTI_SK_CIKLAS1:
		cmp		ax, 0
		je		RASYTI_SK_CIKLAS2
		xor		dx, dx
		div		daliklis		; ax - dalmuo, dx - liekana, daliklis=10
		push	dx
		inc		cx
		jmp		RASYTI_SK_CIKLAS1
	
		; skaitmenys nuskaitomi iš steko ir įrašomi į buferį
	RASYTI_SK_CIKLAS2:
		cmp		cx, 0
		je		BAIGTI_RASYTI_SK
		pop		dx
		add		dl, '0'
		mov		byte ptr [di], dl
		inc		di
		dec		cx
		jmp		RASYTI_SK_CIKLAS2
		
	BAIGTI_RASYTI_SK:
		pop		dx
		pop		cx
		ret
	RASYTI_SK endp
	
CSeg ends
;******************************************************************************

end MAIN

